import QtQuick 1.0
import MapsPlugin 1.0

import "MapApi.js" as MapApi

Item {
    id: mapBehavior

    property variant mapView

    Persistence {
        id: db

        Component.onDestruction: {
            internals.doSave();
        }
    }

    Item {
        id: internals

        property string _mapRevision: "map_rev_latest"
        property bool firstMapRunEver: false
        property bool triggerLayersUiReset: false
        property bool initialCenterChangeMode: true
        property int _previousZoomLevel: -1
        property bool applicationMaximized: true

        function conditionalQuickSave() {
            if (internals.firstMapRunEver) {
                internals.doSave();
                internals.firstMapRunEver = false;
            }
        }

        function doSave() {
            internals.saveMapState();
        }

        function conditionalClearAll() {
            if (MapApi.MapLayers.isPublicTransportMode() || MapApi.MapLayers.isPublicTransportModePartial()) {
                clearAll();
            }
        }

        function clearAll() {
            mapView.clearTransitHighlights();
            mapView.mapSharedData.currentTransitStopInfo = null;
            mapView.mapSharedData.currentTransitStopInfoLines = null;
            mapView.mapSharedData.publicTransportLineHightlighted = false;

            if (MapApi.MapLayers.isPublicTransportModePartial()) {
                MapApi.MapLayers.setPublicTransportModePartialOff();
            }

            if (mapView.model.selectedObject || mapView.model.pureSelectedObject) {
                mapView.model.selectObject(null, false);
            }
        }

        function applyCenter(center) {
            center = center || mapView.mapSharedData.currentCenter;

            if (center && center.latitude && center.longitude) {
                if (mapView.center.latitude != center.latitude || mapView.center.longitude != center.longitude) {
                    mapView.center.latitude = center.latitude;
                    mapView.center.longitude = center.longitude;
                    mapView.setControl("close-infobubble");
                }
            }
        }

        function applyMapSettingsFromMapGlobalData() {
            if (mapView.mapGlobalData) {
                mapView.currentPosition = mapView.mapGlobalData.currentPosition;
                mapView.previousPosition = mapView.mapGlobalData.previousPosition;
                mapView.positionType = mapView.mapGlobalData.currentPositionType;
            }
        }

        function applyMapSettingsFromMapSharedData() {
            mapView.currentPosition = mapView.mapSharedData.currentPosition;
            mapView.previousPosition = mapView.mapSharedData.previousPosition;
            mapView.positionType = mapView.mapSharedData.currentPositionType;

            if (mapView.mapSharedData.currentZoomScale) {
                MapApi.zoomTo(mapView.mapSharedData.currentZoomScale, null, true);
            }

            if (mapView.mapSharedData.currentDrawnHaloSize !== null) {
                mapView.forceChangeAccuracy(mapView.mapSharedData.currentDrawnHaloSize);
            }

            if (mapView.mapSharedData.currentTransitStopInfo && mapView.mapSharedData.currentTransitStopInfo.lines) {
                mapView.highlightTransitLines(mapView.mapSharedData.currentTransitStopInfo.lines);
            } else {
                mapView.clearTransitHighlights();
            }

            mapView.attachedToLocation = mapView.mapSharedData.currentAttachedToLocation;

            if (mainscreen.enableAutomaticRotating) {
                mapView.automaticRotating = true;
            } else {
                mapView.automaticRotating = false;
            }
        }

        function checkIfOrientationShallBeFixed(customZoomScale) {
            return !mapView.compassMode && !mapView.previousCompassMode && !appState.inGuidanceMode && (internals._previousZoomLevel <= 2 || !mainscreen.enableAutomaticRotating) && MapApi.getCurrentZoomLevel(customZoomScale) >= 2 && mapView.orientation != 0;
        }

        function adjustOrientation() {
            if (internals.checkIfOrientationShallBeFixed()) {
                mapView.moveTo(
                   null,
                   Map.ANIMATION_LINEAR,
                   Map.PRESERVE_ZOOM_SCALE,
                   0,
                   Map.PRESERVE_PERSPECTIVE
                );
            }
        }

        function adjustKinetics() {
            var kineticHalfLifes = mapView.kineticsHalfLifes;
            var currentZoomLevel = MapApi.getCurrentZoomLevel();
            if (kineticHalfLifes[currentZoomLevel]) {
                mapView.kineticPanning.halfLife = kineticHalfLifes[currentZoomLevel];
            } else {
                mapView.kineticPanning.halfLife = mapView.kineticsHalfLife;
            }
        }

        function saveMapState() {
            if (!mapView.dontSaveData) {
                var data = mapView.mapSharedData;

                if (data.activeMap == mapView) {
                    if (data.currentCenter) {
                        var data = {
                            "latitude": data.currentCenter? data.currentCenter.latitude : 0,
                            "longitude": data.currentCenter? data.currentCenter.longitude : 0,
                            "currentPositionLatitude": data.currentPosition? data.currentPosition.geoCoordinates.latitude : 0,
                            "currentPositionLongitude": data.currentPosition? data.currentPosition.geoCoordinates.longitude : 0,
                            "previousPositionLatitude": data.previousPosition? data.previousPosition.geoCoordinates.latitude : 0,
                            "previousPositionLongitude": data.previousPosition? data.previousPosition.geoCoordinates.longitude : 0,
                            "zoomScale": data.currentZoomScale,
                            "timeMode": data.timeMode,
                            "mapMode": data.mapMode,
                            "mode3d": data.mode3d,
                            "landmarksVisible": data.landmarksVisible,
                            "trafficInfoVisible": data.trafficInfoVisible,
                            "publicTransportMode": data.publicTransportMode,
                            "publicTransportModePartial": data.publicTransportModePartial,
                            "publicTransportLineHightlighted": data.publicTransportLineHightlighted
                        };

                        var tmpSkipPropertiesOnSaveTime = mapView.mapSharedData.skipPropertiesOnSaveTime;
                        if (tmpSkipPropertiesOnSaveTime) {
                            for (var i = 0; i < tmpSkipPropertiesOnSaveTime.length; i++) {
                                delete data[tmpSkipPropertiesOnSaveTime[i]];
                            }
                        }

                        if (!data.previousPositionLatitude) {
                            delete data.previousPositionLatitude;
                        }
                        if (!data.previousPositionLongitude) {
                            delete data.previousPositionLongitude;
                        }

                        db.setRevision(data, _mapRevision);
                    }
                }
            }
        }

        function setMapDefaultStates(mode) {
            if (mode == "no-position-at-all") {
                mapView.center.latitude = 45.969479;
                mapView.center.longitude = 15.717772;
                mapView.zoomScale = 4530902;

                mapView.mode = mapView._NO_POSITION_AT_ALL_;

                internals.firstMapRunEver = true;
            }
        }

        function loadMapState() {
            if (!mapView.getBehavior(mapView._NO_INITIAL_SETTINGS_LOAD_)) {
                if (mapView.mapSharedData.state == "default" && mapView.mapSharedData.mode == "default") {
                    db.getRevision(_mapRevision, function(rs) {
                        if (rs) {
                            var latitude = parseFloat(rs["latitude"]);
                            var longitude = parseFloat(rs["longitude"]);
                            var currentPositionLatitude = parseFloat(rs["currentPositionLatitude"]);
                            var currentPositionLongitude = parseFloat(rs["currentPositionLongitude"]);
                            var previousPositionLatitude = parseFloat(rs["previousPositionLatitude"]);
                            var previousPositionLongitude = parseFloat(rs["previousPositionLongitude"]);
                            var latitude = parseFloat(rs["latitude"]);
                            var longitude = parseFloat(rs["longitude"]);
                            var zoomScale = parseInt(rs["zoomScale"]);
                            var timeMode = rs["timeMode"];
                            var mapMode = rs["mapMode"];
                            var mode3d = rs["mode3d"] == "true" ? true : false;
                            var landmarksVisible = rs["landmarksVisible"] == "true" ? true : false;

                            var trafficInfoVisible = rs["trafficInfoVisible"] == "true" ? true : false;
                            // var trafficInfoVisible = false;

                            var publicTransportMode = rs["publicTransportMode"] == "true" ? true : false;
                            //var publicTransportModePartial = rs["publicTransportModePartial"] == "true" ? true : false;
                            var publicTransportModePartial = false
                            //var publicTransportLineHightlighted = rs["publicTransportLineHightlighted"] == "true" ? true : false;
                            var publicTransportLineHightlighted = false;

                            if (currentPositionLatitude && currentPositionLongitude) {
                                delayedOldPositionApplier.go(currentPositionLatitude, currentPositionLongitude);
                            } else if (previousPositionLatitude && previousPositionLongitude) {
                                delayedOldPositionApplier.go(previousPositionLatitude, previousPositionLongitude);
                            }

                            mapView.mapSharedData.state = "modified";

                            mapView.mapSharedData.currentCenter = mapView.center;
                            if (!mapView.startupCenterLatitude && !mapView.startupCenterLongitude) {
                                mapView.mapSharedData.currentCenter.latitude = latitude;
                                mapView.mapSharedData.currentCenter.longitude = longitude;
                                mapView.mapSharedData.currentZoomScale = zoomScale;
                            }

                            mapView.mapSharedData.timeMode = timeMode;
                            //mapView.mapSharedData.mapMode = mapMode;
                            mapView.mapSharedData.mode3d = mode3d;
                            mapView.mapSharedData.landmarksVisible = landmarksVisible;
                            mapView.mapSharedData.trafficInfoVisible = trafficInfoVisible;
                            mapView.mapSharedData.publicTransportMode = publicTransportMode;
                            mapView.mapSharedData.publicTransportModePartial = publicTransportModePartial;
                            mapView.mapSharedData.publicTransportLineHightlighted = publicTransportLineHightlighted;

                            mapView.mapSharedData.mode = "loaded";
                        } else {
                            setMapDefaultStates("no-position-at-all");
                        }
                    });
                }
            }
        }

        property bool restorePositionProviderToStarted: false
        function conditionalStopPositionProvider() {
            if (mapView.positionProviderEnabled && !appState.inGuidanceMode) {
                mapView.stopPositionProvider();
                internals.restorePositionProviderToStarted = true;
            }
        }
        function conditionalStartPositionProvider() {
            if (internals.restorePositionProviderToStarted) {
                mapView.startPositionProvider();
                internals.restorePositionProviderToStarted = false;
            }
        }

        Timer {
            id: delayedOldPositionApplier

            interval: 1 * 1000
            running: false

            property variant latitude: null
            property variant longitude: null

            function go(latitude, longitude) {
                delayedOldPositionApplier.latitude = latitude;
                delayedOldPositionApplier.longitude = longitude;
                start();
            }

            onTriggered: {
                if (!mapView.previousPosition) {
                    var geoCoordinates = geoCoordinatesFactory.createObject(mapView);
                    geoCoordinates.latitude = delayedOldPositionApplier.latitude;
                    geoCoordinates.longitude = delayedOldPositionApplier.longitude;
                    mapView.previousPosition = {fake: true, geoCoordinates: geoCoordinates};
                }
            }
        }

        Component {
            id: geoCoordinatesFactory
            GeoCoordinates {}
        }
    }

    Timer {
        id: lazyZoomInDoubleClicked
        interval: 1
        onTriggered: { MapApi.zoomIn(spaceGeoCoordinates, internals.checkIfOrientationShallBeFixed(mapView.zoomScale / MapApi.ZOOM_STEP_MULTIPLIER)? 0 : undefined); }
    }

    Timer {
        id: lazyZoomOutOnTwoFingers
        interval: 1
        onTriggered: { MapApi.zoomOut(spaceGeoCoordinates); }
    }

    GeoCoordinates { id: spaceGeoCoordinates }

    Connections {
        target: mapView
        ignoreUnknownSignals: true

        onMapDoubleClicked: {
            spaceGeoCoordinates.latitude = geoPos.latitude;
            spaceGeoCoordinates.longitude = geoPos.longitude;
            lazyZoomInDoubleClicked.start();
        }

        onMapTwoFingerDoubleClicked: {
            spaceGeoCoordinates.latitude = geoPos.latitude;
            spaceGeoCoordinates.longitude = geoPos.longitude;
            lazyZoomOutOnTwoFingers.start();
        }

        onTwoFingerSlide: {
            if (mainscreen.enableMagic3d2d) {
                if (deltaY) {
                    if (mapView.perspective + deltaY < 50 && mapView.perspective + deltaY > 0) {
                        mapView.perspective += deltaY;
                    }
                }
            }
        }

        onOrientationChanged: {
            if (mainscreen.enableAutomaticRotating || mainscreen.magicCompass) {
                if (mapView.mapSharedData.orientation != mapView.orientation) {
                    mapView.mapSharedData.orientation = mapView.orientation;
                }
            }
        }

        onMapObjectsSelected: {
            mapView.mapSharedData.currentMapObjectSelected = mapObject;

            if (!MapApi.MapLayers.isPublicTransportModeActive() && (mapObject.type == MapObject.TRANSITSTOP)) {
                MapApi.MapLayers.setPublicTransportModePartialOn();
            }

            if (mapObject && mapObject.type == MapObject.TRANSITSTOP && (MapApi.MapLayers.isPublicTransportMode() || MapApi.MapLayers.isPublicTransportModePartial())) {
                mapView.clearTransitHighlights();
                mapView.mapSharedData.currentTransitStopInfo = mapView.mapSharedData.currentMapObjectSelected.transitStopInfo;

                mapView.mapSharedData.currentTransitStopInfoLines = null;
                if (mapView.mapSharedData.currentTransitStopInfo.lines.length > 0) {
                    mapView.mapSharedData.currentTransitStopInfoLines = mapView.mapSharedData.currentTransitStopInfo.lines;

                    mapView.highlightTransitLines(mapView.mapSharedData.currentTransitStopInfoLines);
                    mapView.mapSharedData.publicTransportLineHightlighted = true;
                } else {
                    mapView.mapSharedData.publicTransportModePartial = false;
                    MapApi.MapLayers.setPublicTransportModePartialOff();
                    MapApi.MapLayers.reset();
                }
            }
        }

        onControlChanged: {
            if (mapView.control == "clear-transit-hightlights") {
                mapView.clearTransitHighlights();
                mapView.mapSharedData.currentTransitStopInfo = null;
                mapView.mapSharedData.currentTransitStopInfoLines = null;
                mapView.mapSharedData.publicTransportLineHightlighted = false;
            } else if (mapView.control == "conditional-clear-all") {
                internals.conditionalClearAll();
            } else if (mapView.control == "force-close-infobubble") {
                mapView.closeInfoBubbleByForce();
            } else if (mapView.control == "emit-force-close-infobubble") {
                mapView.mapSharedData.setControl("force-close-infobubble");
            } else if (mapView.control == "emit-turn-on-compass") {
                mapView.mapSharedData.setControl("turn-on-compass");
            } else if (mapView.control == "emit-turn-off-compass") {
                mapView.mapSharedData.setControl("turn-off-compass");
            } else if (mapView.control == "emit-forget-selected-marker") {
                mapView.mapSharedData.setControl("forget-selected-marker");
            } else if (mapView.control == "ui-changed") {
                MapApi.MapLayers.reset();
            } else if (mapView.control == "emit-ui-changed") {
                mapView.mapSharedData.setControl("ui-changed");
            }
        }

        onMapClicked: {
            mapView.mapSharedData.currentMapObjectSelected = mapObject;
            if (!mapObject || mapObject.type != MapObject.TRANSITSTOP) {
                internals.conditionalClearAll();
            }
        }

        onMapDragged: {
            //internals.conditionalClearAll();
            mapView.mapSharedData.state = "dragged";
        }

        onMapLongPressed: {
            internals.conditionalClearAll();
            mapView.mapSharedData.state = "longPressed";
        }

        onZoomScaleChanged: {
            if (!mainscreen.enableMagic3d2d) {
                mapView.mapSharedData.currentZoomScale = mapView.zoomScale;
            }
            internals.adjustOrientation();
            internals.adjustKinetics();

            internals._previousZoomLevel = MapApi.getCurrentZoomLevel();
        }

        onAnimationDone: {
            mapView.mapSharedData.currentZoomScale = mapView.zoomScale;

            internals.conditionalQuickSave();
        }

        onVisibleChanged: {
            if (mapView.visible) {
                internals.applyMapSettingsFromMapSharedData();
                internals.applyMapSettingsFromMapGlobalData();
                internals.applyCenter(mapView.mapSharedData.currentCenter);
                if (internals.triggerLayersUiReset) {
                    MapApi.MapLayers.reset();
                    internals.triggerLayersUiReset = false;
                }

                if (mainscreen.enableMagic3d2d) {
                    mapView.perspective = mapView.mapSharedData.perspective;
                }
                if (mainscreen.enableAutomaticRotating) {
                    mapView.orientation = mapView.mapSharedData.orientation;
                }

                mapView.mapSharedData.activeMap = mapView;
                if (!mapView.dontPopulateGlobalActiveMap) {
                    mapView.mapGlobalData.activeMap = mapView;
                }
            } else {
                mapView.status = mapView._INTERRUPTED_ANIMATION_;
            }
        }

        onCenterChanged: {
            mapView.mapSharedData.updateCurrentCenter(mapView.center);

            if (internals.initialCenterChangeMode) {
                internals.initialCenterChangeMode = false;
            } else {
                mapView.mapSharedData.state = "center-changed";
            }
        }

        onCurrentPositionChanged: {
            mapView.mapSharedData.updateCurrentPosition(mapView.currentPosition);
            mapView.mapGlobalData.updateCurrentPosition(mapView.currentPosition);
        }

        onPreviousPositionChanged: {
            mapView.mapSharedData.updatePreviousPosition(mapView.previousPosition);
            mapView.mapGlobalData.updatePreviousPosition(mapView.previousPosition);
        }

        onPositionTypeChanged: {
            mapView.mapSharedData.updateCurrentPositionType(mapView.positionType);
            mapView.mapGlobalData.updateCurrentPositionType(mapView.positionType);
        }

        onDrawnHaloSizeChanged: {
            mapView.mapSharedData.currentDrawnHaloSize = mapView.drawnHaloSize;
        }

        onAttachedToLocationChanged: {
            mapView.mapSharedData.currentAttachedToLocation = mapView.attachedToLocation;
        }

        onPerspectiveChanged: {
            if (!mapView.preventAutomaticPropertyFiltering) {
                MapApi.MapLayers.tryPerspective(mapView.perspective);
            }
            if (mainscreen.enableMagic3d2d) {
                mapView.mapSharedData.perspective = mapView.perspective;
            }
        }

        onSchemeChanged: {
            if (!mapView.preventAutomaticPropertyFiltering) {
                MapApi.MapLayers.tryScheme(mapView.scheme);
            }
        }

        onLandmarksVisibleChanged: {
            if (!mapView.preventAutomaticPropertyFiltering) {
                MapApi.MapLayers.tryLandmakrsVisible(mapView.landmarksVisible);
            }
        }

        onTrafficInfoVisibleChanged: {
            if (!mapView.preventAutomaticPropertyFiltering) {
                MapApi.MapLayers.tryTrafficModeVisible(mapView.trafficInfoVisible);
            }
        }

        onTransitModeChanged: {
            if (!mapView.preventAutomaticPropertyFiltering) {
                MapApi.MapLayers.tryPublicTransportMode(mapView.transitMode);
            }
        }

        onKineticsEnabledChanged: {
            mapView.kineticPanning.enabled = mapView.kineticsEnabled;
        }

        onBlockInteractivityChanged: {
            if (mapView.blockInteractivity) {
                mapView.automaticPanning = false;
                mapView.automaticRotating = false;
                mapView.automaticZooming = false;
            } else {
                mapView.automaticPanning = true;
                mapView.automaticRotating = mainscreen.enableAutomaticRotating;
                mapView.automaticZooming = true;
            }
        }

        onInvisibleBlockInteractivityChanged: {
            if (mapView.invisibleBlockInteractivity) {
                mapView.automaticPanning = false;
                mapView.automaticRotating = false;
                mapView.automaticZooming = false;
            } else {
                mapView.automaticPanning = true;
                mapView.automaticRotating = mainscreen.enableAutomaticRotating;
                mapView.automaticZooming = true;
            }
        }

        onBehaviorsModified: {
            if (key == mapView._NO_INTERACTIVITY_) {
                if (value) {
                    mapView.automaticPanning = false;
                    mapView.automaticRotating = false;
                    mapView.automaticZooming = false;
                } else {
                    mapView.automaticPanning = true;
                    mapView.automaticRotating = mainscreen.enableAutomaticRotating;
                    mapView.automaticZooming = true;
                }
            }
        }

        onLazyAppendIconModeChanged: {
            if (mapView.lazyAppendIconMode) {
                autoStopLazyAppendIconMode.start();
            } else {
                autoStopLazyAppendIconMode.stop();
            }
        }

        Component.onCompleted: {
            if (!mapView.mainScreen) {
                var rootElement = MapApi.findRootApplicationLevel(mapView);
                if (rootElement != mapView) {
                    mapView.mainScreen = rootElement;
                } else {
                    mapView.mainScreen = parent;
                }
            }

            mapView.mainScreen = mapView.mainScreen? mapView.mainScreen : MapApi.findRootApplicationLevel(mapView);
            if (mapView.mainScreen && mapView.mainScreen.mapSharedData) {
                mapView.mapSharedData = mapView.mainScreen.mapSharedData;
            }
            if (mainscreen && mainscreen.mapGlobalData) {
                mapView.mapGlobalData = mainscreen.mapGlobalData;
            }

            internals.loadMapState();

            if (!mapView.mapSharedData.currentCenter) {
                mapView.mapSharedData.updateCurrentCenter(mapView.center);
            }

            internals.applyMapSettingsFromMapSharedData();

            if (MapApi.MapLayers) {
                MapApi.MapLayers.reset();
                MapApi.setCategory("AIRPORT", true);
            }

            //mapView.mouseMoveTolerance = 10;

            mapView.kineticPanning.enabled = mapView.kineticsEnabled;
            internals.adjustKinetics();

            PositionObject.visible = false;

            internals.applyCenter(null);

            if (mapView.mapGlobalData.currentPosition) {
                mapView.currentPosition = mapView.mapGlobalData.currentPosition;
            }
            if (mapView.mapGlobalData.previousPosition) {
                mapView.previousPosition = mapView.mapGlobalData.previousPosition;
            }

            mapView.ready = true;

            if (mapView.visible) {
                mapView.mapSharedData.activeMap = mapView;
                if (!mapView.dontPopulateGlobalActiveMap) {
                    mapView.mapGlobalData.activeMap = mapView;
                }
            }

            if (mainscreen.magicAccelerometer) {
                log.info("Will start the accelerometer...");
                appState.startAccelerometer();
            }

            if (typeof(rootApplicationLevel) == "object" && rootApplicationLevel.mapLocked) {
                mapView.blockInteractivity = true;
            }
        }

        Component.onDestruction: {
            if (mapView.mapSharedData.activeMap == mapView) {
                mapView.mapSharedData.activeMap = null;
            }
            if (mapView.mapGlobalData.activeMap == mapView) {
                mapView.mapGlobalData.activeMap = null;
            }
        }
    }

    Connections {
        target: rootApplicationLevel
        ignoreUnknownSignals: true

        onSignInComplete: {
            if (success) {
                mapView.blockInteractivity = false;
            }
        }
    }

    Connections {
        target: mapView.model
        ignoreUnknownSignals: true

        onPureSelectedObjectChanged: {
            if (!mapView.model.pureSelectedObject) {
                //internals.conditionalClearAll();
            }
        }
    }

    Connections {
        target: mainscreen
        ignoreUnknownSignals: true

        onEnableAutomaticRotatingChanged: {
            mapView.automaticRotating = mainscreen.enableAutomaticRotating;
        }
    }

    Connections {
        target: mapView.mapSharedData
        ignoreUnknownSignals: true

        onUiChanged: {
            internals.triggerLayersUiReset = true;
        }

        onOrientationChanged: {
            if (mainscreen.enableAutomaticRotating || mainscreen.magicCompass) {
                mapView.orientation = mapView.mapSharedData.orientation;
            }
        }

        onControlChanged: {
            mapView.setControl(mapView.mapSharedData.control);
        }

        onPropagandaChanged: {
            mapView.propaganda = mapView.mapSharedData.propaganda;
        }
    }

    Connections {
        target: appState
        ignoreUnknownSignals: true

        onMinimized: {
        }
        onMaximized: {
        }
        onWindowDeactivated: {
            log.trace(">>>>>>>>>>>>>>>>>>>>                           WILL STOP POSITION PROVIDER IN 2 MINUTES...........................")
            lazyPositionStopper.start();
        }
        onWindowActivated: {
            log.trace(">>>>>>>>>>>>>>>>>>>>                           WILL START POSITION PROVIDER NOW!!!!")
            lazyPositionStopper.stop();
            internals.conditionalStartPositionProvider();
        }

        onAccelerometerReading: {
            if (mainscreen.magicAccelerometer) {
                if (mapView.visible) {
                    var valueToMonitorFlip;
                    var valueToMonitorSwirl;
                    var signOfSwirl;
                    if (mainscreen.state == "portrait") {
                        valueToMonitorFlip = y;
                        valueToMonitorSwirl = x;
                        signOfSwirl = -1;
                    } else {
                        signOfSwirl = 1;
                        valueToMonitorFlip = x;
                        valueToMonitorSwirl = y;
                    }

                    var tmpFlip = Math.abs(valueToMonitorFlip) - 4;
                    tmpFlip = tmpFlip < 0? 0 : tmpFlip;
                    tmpFlip = tmpFlip > 4? 4 : tmpFlip;

                    var tmpSwirlSign = valueToMonitorSwirl < 0? signOfSwirl : -signOfSwirl;
                    var tmpSwirl = Math.abs(valueToMonitorSwirl);
                    tmpSwirl = tmpSwirl < 0? 0 : tmpSwirl;
                    tmpSwirl = tmpSwirl > 3? 3 : tmpSwirl;



                    var percentsOfFlip = tmpFlip / 4;
                    var percentsOfSwirl = tmpSwirl / 3 * tmpSwirlSign;

                    var newPerspective = 55 - 55 * (1 - percentsOfFlip);
                    var newOrientation = 65 - 65 * percentsOfSwirl;

                    mapView.orientation = newOrientation;
                    mapView.perspective = newPerspective;

                    /*
                    mapView.moveTo(
                        null,
                        Map.ANIMATION_LINEAR,
                        Map.PRESERVE_ZOOM_SCALE,
                        newOrientation,
                        newPerspective
                    );
                    */
                }
            }
        }
    }

    Connections {
        target: appState
        ignoreUnknownSignals: true

        onMinimized: { }
        onMaximized: { }
        onWindowDeactivated: {
            internals.applicationMaximized = false;
        }
        onWindowActivated: {
            internals.applicationMaximized = true;
        }
    }

    Timer {
        id: lazyPositionStopper
        interval: 1000 * 60 * 2

        onTriggered: {
            log.trace(">>>>>>>>>>>>>>>>>>>>                           WILL STOP POSITION PROVIDER NOW!!!!")
            internals.conditionalStopPositionProvider();
        }
    }

    Timer {
        id: autoStopLazyAppendIconMode
        interval: 1000 * 2
        onTriggered: { mapView.lazyAppendIconMode = false; }
    }

    Timer {
        id: autoSaveAllSettingsToStorageJustInCasePhoneCrashesAndWeCannotCallDestructor
        interval: 1000 * 60 * 15
        repeat: true
        running: mapView.visible && internals.applicationMaximized
        onTriggered: {
            log.info("Will save the map settings just in case the whole phone crashes and we cannot reach destructor...");
            internals.doSave();
        }
    }
}
